import unittest

from silica_django.config import SilicaConfig, SilicaFieldConfig
from silica_django.utils.jsonschema import JsonSchemaUtils

from silica_django.rules import Or, And, Not, ShowIf


class BaseTestCase(unittest.TestCase):
    # set maximum number of characters which will be printed for a single diff
    maxDiff = 10000

    def assertEqualAsStrings(self, first, second, msg=None):
        """ Converts arguments to strings before comparing them; helps with legibility when reading printout from 
        comparing heavily nested objects """
        return self.assertEqual(str(first), str(second), msg=msg)


class TestConditions(BaseTestCase):
    def test_or_single_key_single_value(self):
        or_1 = Or(key1=1)
        self.assertEqualAsStrings(or_1.get_condition_schema(), {
            "anyOf": [
                {"type": "object", "properties": {"key1": {"const": 1}}}
            ]
        })

    def test_or_single_key_multiple_values(self):
        or_1 = Or(key1=[1, 2, 3])
        self.assertEqualAsStrings(or_1.get_condition_schema(), {
            "anyOf": [
                {"type": "object",
                 "properties": {"key1": {"enum": [1, 2, 3]}}}
            ]
        })

    def test_or_multiple_keys(self):
        or_1 = Or(key1=1, key2=[1, 2, 3])
        self.assertEqualAsStrings(or_1.get_condition_schema(), {
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"key1": {"const": 1}, "key2": {"enum": [1, 2, 3]}}
                }
            ]
        })

    def test_and_multiple_keys(self):
        and_1 = And(key1=1, key2=2)
        self.assertEqualAsStrings(and_1.get_condition_schema(), {
            "allOf": [
                {
                    "type": "object",
                    "properties": {
                        "key1": {"const": 1},
                        "key2": {"const": 2}
                    }
                }
            ]
        })

    def test_not(self):
        not_1 = Not(key1=1)
        self.assertEqualAsStrings(not_1.get_condition_schema(), {
            'not': {
                'type': 'object',
                'properties':
                    {
                        'key1': {'const': 1}
                    }
            }
        })

    def test_not_composable(self):
        not_1 = Not(Or(key1=1, key2=1))
        self.assertEqualAsStrings(not_1.get_condition_schema(), {
            'not': {
                'anyOf': [
                    {
                        'type': 'object',
                        'properties': {
                            'key1': {'const': 1},
                            'key2': {'const': 1},
                        }
                    }
                ]
            }
        })


class TestRules(BaseTestCase):
    def test_show_only_kwargs(self):
        rule = ShowIf(key1=1)
        self.assertEqualAsStrings(rule.get_schema(), {
            "effect": rule.effect,
            "condition": {
                "scope": "#",
                "schema": {
                    "anyOf": [
                        {
                            "allOf": [
                                {
                                    "type": "object",
                                    "properties": {
                                        "key1": {"const": 1}
                                    }
                                }
                            ]
                        }
                    ]
                }
            }
        })

    def test_show_only_args(self):
        rule = ShowIf(Or(key1=1, key2=1), And(key1=2, key2=3))
        self.assertEqualAsStrings(rule.get_schema(), {
            "effect": rule.effect,
            "condition": {
                "scope": "#",
                "schema": {
                    "anyOf": [
                        {
                            "anyOf": [
                                {
                                    "type": "object",
                                    "properties": {
                                        "key1": {"const": 1},
                                        "key2": {"const": 1},
                                    }
                                }
                            ]
                        },
                        {
                            "allOf": [
                                {
                                    "type": "object",
                                    "properties": {
                                        "key1": {"const": 2},
                                        "key2": {"const": 3},
                                    }
                                }
                            ]
                        }
                    ]
                }
            }
        })

    def test_show_args_and_kwargs(self):
        rule = ShowIf(And(key2=2, key3=3), key1=1)
        self.assertEqualAsStrings(rule.get_schema(), {
            "effect": rule.effect,
            "condition": {
                "scope": "#",
                "schema": {
                    "anyOf": [
                        {
                            "allOf": [
                                {
                                    "type": "object",
                                    "properties": {
                                        "key2": {"const": 2},
                                        "key3": {"const": 3},
                                    }
                                }
                            ]
                        },
                        {
                            "allOf": [
                                {
                                    "type": "object",
                                    "properties": {
                                        "key1": {"const": 1}
                                    }
                                }
                            ]
                        },
                    ]
                }
            }
        })


class TestJsonSchemaTranslation(BaseTestCase):

    def test_value_as_jsonschema(self):
        self.assertEqualAsStrings(JsonSchemaUtils.value_as_jsonschema(1), {"const": 1})
        self.assertEqualAsStrings(JsonSchemaUtils.value_as_jsonschema([1]), {"enum": [1]})


class TestSilicaConfig(BaseTestCase):
    def assertKwargsCorrectlyProcessed(self):
        """ Assert that the kwargs passed to the SilicaConfig are properly formatted internally """
        config = SilicaConfig(
            field1=SilicaFieldConfig(
                maximum=1,
                minimum=2,
                default=2,
                min_length=1,
                max_length=3,
                description="a description",
                type="object",
                schema_format="int",
                required=["field2"],
                additional_properties=True,
                label="oh yeah",
            )
        )
        
        
        return True
    
    def assertUISchemaCustomize(self):
        return True
    
    def assertUISchemaOverride(self):
        return True
    
    def assertSchemaCustomize(self):
        return True
    
    def assertSchemaOverride(self):
        return True
    
    def assertRule(self):
        return True
    
    def assertComplexUISchema(self):
        return True

if __name__ == "__main__":
    unittest.main()



from django import forms

from silica_django import fields
from silica_django.fields import SilicaSubFormArrayField, SilicaSubmitInputField
from silica_django.utils.jsonschema import JsonSchemaUtils
from silica_django.widgets import SilicaRenderer


class JsonSchemaMixin(JsonSchemaUtils):
    """ Contains utility functions for interfacing between native python/django and jsonschema """
    def _django_to_jsonschema_field(self, field_name, field, field_config=None):
        # most field types are string by default
        field_type = "string"
        # format is only required for some special types e.g. date
        field_kwargs = {
            'name': field_name,
            'options': {}
        }
        if isinstance(field, forms.DateField):
            field_kwargs["format"] = "date"
        elif isinstance(field, forms.DateTimeField):
            field_kwargs["format"] = "date-time"
        elif isinstance(field, forms.TimeField):
            field_kwargs["format"] = "time"
        elif isinstance(field, forms.IntegerField):
            field_type = "integer"
        elif isinstance(field, forms.FloatField) or isinstance(field, forms.DecimalField):
            field_type = "number"
        elif isinstance(field, forms.BooleanField):
            field_type = "boolean"
        # todo: differentiate between arrays of related items and a multi field (e.g. tags)
        elif isinstance(field, fields.SilicaSubFormArrayField):
            field_type = "array"
            if field._instantiated_forms:
                item_schema = field._instantiated_forms[0].get_schema()
            else:
                # there are no existing sub-items, instantiate the form to get the schema
                item_schema = field.instance_form().get_schema()
            item_schema["properties"][field.identifier_field] = {
                    "type": "number",
                    "hidden": True
                }
            field_kwargs['items'] = {
                **item_schema,
            }
        if hasattr(field, 'choices'):
            field_kwargs["oneOf"] = [{'const': value, 'title': title} for (value, title) in field.choices]
        # special checks
        if isinstance(field.widget, forms.HiddenInput):
            field_kwargs['hidden'] = True
        if field.disabled:
            field_kwargs['readOnly'] = True
        if isinstance(field.widget, forms.RadioSelect):
            field_kwargs['options']['format'] = "radio"
            # for a radio select, everything is a string - we'll convert on the backend
            field_type = "string"
            if not hasattr(field, 'choices'):
                field_kwargs["oneOf"] = [{'const': str(value), 'title': title} for (value, title) in field.widget.choices]
        if field_config:
            if field_config.schema:
                field_kwargs.update(field_config.schema)
        if isinstance(field, SilicaSubFormArrayField):
            if field.min_instances:
                field_kwargs['minItems'] = field.min_instances
            if field.max_instances:
                field_kwargs['maxItems'] = field.max_instances
        return {
            "type": field_type,
            **field_kwargs
        }

    def _django_widget_to_ui_schema(self, field, field_config=None):
        ui_schema = {
            'options': {}
        }
        if field.label:
            ui_schema['label'] = field.label
        if field.disabled:
            ui_schema['readonly'] = True
        # special values for widgets
        if isinstance(field.widget, forms.Textarea):
            ui_schema['options']['multi'] = True
        if isinstance(field.widget, SilicaRenderer):
            ui_schema['options']['customComponentName'] = field.widget.custom_component_name
        if isinstance(field, SilicaSubmitInputField):
            ui_schema['options']['format'] = 'submit'
        # add rules and update uischema
        if field_config:
            if field_config.rule:
                ui_schema['rule'] = field_config.rule.get_schema()
            if field_config.uischema:
                ui_opts = ui_schema['options']
                ui_opts.update(field_config.uischema['options'])
                ui_schema.update(field_config.uischema)
                ui_schema['options'] = ui_opts
        return ui_schema



from django import template

register = template.Library()


@register.inclusion_tag('silica_loader.html')
def load_silica_form(form, form_id):
    return {
        "form": form,
        "form_data_key": form_id + "-data",
        "form_schema_key": form_id + "-schema",
        "form_ui_schema_key": form_id + "-ui-schema",
        "form_errors_key": form_id + "-errors",
    }



from typing import Dict, Any


class JsonSchemaUtils:
    @staticmethod
    def value_as_jsonschema(value) -> Dict[str, Any]:
        if type(value) == list:
            return {"enum": value}
        else:
            return {"const": value}
